#ifndef _IDENTITY_H
#define _IDENTITY_H

#include "DistributionLibrairie.h"
#include "IdentityIdl.h"

using namespace std;

////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////

class Identity : public DistributionLibrary
{
private:
  ////////////////////////////////////////
  // Client & Server side

  unsigned _config; // vector of bit describing which elements have been configured (cf .cc)

  PaCO::PacoTopology_t _sTopo; // topology of the source (sender)  -- SET
  PaCO::PacoTopology_t _dTopo; // topology of the destination (receiver) -- SET

  unsigned _nodeRank; // id of the node (either source or destination) -- SET
  unsigned _llen;     // local vector size: in element -- SET
  unsigned _unitsize; // #bytes of one element -- SET

  ////////////////////////////////////////
  // Client Side
  char* _clientBuffer; // pointer to the user data (set)
  IdentityLib::IdentityDescr _clientDescr; // contains the descr exchange between sender & destination -- SET indirectly

  ////////////////////////////////////////
  // Server Side
  unsigned long _serverToReceived; // #of element yet to be received (during an invocation)
  char* _serverBuffer; // Buffer where to receive the data

  IdentityLib::IdentityDescr* _serverDescr; // contains the descr exchange between sender & destination -- SET indirectly

  ////////////////////////////////////////////////////////////
  ////////////////////////////////////////////////////////////

public:
  Identity();
  virtual ~Identity();

  // Common
  virtual void setSourceTopology(PaCO::PacoTopology_t topo); 
  virtual PaCO::PacoTopology_t getSourceTopology(); 
  virtual void setDestTopology(PaCO::PacoTopology_t topo); 
  virtual PaCO::PacoTopology_t getDestTopology(); 
  virtual void setNodeRank(long Rank);
  virtual long getNodeRank(); 
  virtual void setCommunicator(void * group) {}

  // Data description API
  void setEltSize(unsigned long size);
  void setLocalNbElt(unsigned long elt_nb);

  // Client Side
  virtual PieceToSend* computePiecesToSend(unsigned& size);
  virtual void  setDataPtr(void* dataPtr);
  virtual void* getClientData(void *pid, int server_node, long & size, long & length, bool & end);
  virtual void* getDescr();
  virtual void  clientFree(); // true if com schedule to be freed


  // Server Side
  virtual bool  insertData(void* rcptBuf, unsigned long element_nb);
  virtual void  setDescr(void * descr);

  virtual void* getServerData(long & length);
  virtual void  serverFree();

};

////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////

class Identity_fab : virtual public paco_fabrique_distribution
{

 public:
  Identity_fab() {}
  virtual ~Identity_fab() {}
  
  DistributionLibrary *create()
    {
      return new Identity();
    }

};

#endif
