#ifndef COMMUNICATION_SCHEDULING_IS_DEFINED
#define COMMUNICATION_SCHEDULING_IS_DEFINED

#include "FabriqueManagerTpl.h"
#include "PaCO++_types.h"
#include "PaCO++_operation.h"
#include "com/paco_com.h"

// Abstract class to be implemeted by each scheduling lib

/** paco_comSchedule represents an actual scheduling */
class paco_comSchedule
{
public:

  virtual unsigned int getNbPhases()=0;
  virtual unsigned int getElementPhase(const unsigned phase)=0;
  virtual unsigned int getSrc(const unsigned phase, const unsigned id)=0;
  virtual unsigned int getDst(const unsigned phase, const unsigned id)=0;
  virtual unsigned int getSize(const unsigned phase, const unsigned id)=0;
  virtual        void* getInfo(const unsigned phase, const unsigned id)=0;

};

// Abstract class to be implemeted by each scheduling lib
/** paco_comScheduling is able to compute scheduling, represented by paco_comSchedule */
class paco_comScheduling
{

public:
  paco_comScheduling();
  virtual ~paco_comScheduling();

  virtual paco_comSchedule* computeScheduling(unsigned rank, 
					      PaCO::PacoTopology_t* srctopo, PaCO::PacoTopology_t* dsttopo,
					      unsigned **com_matrix, void** info_vector[])=0;

  /** return NULL if schedule id is unknown */
  paco_comSchedule* getSchedule(long id);

  virtual void setSchedule(long id, paco_comSchedule* cs);

  virtual void clearScheduleId(long id) throw (InvalidArgument);

  virtual void clearAllSchedules();

  virtual void propageComMatrix(PaCO_operation* op, unsigned **comMatrix);
  
  virtual void displayStatus();

protected:
  PaCO::PacoTopology_t* _client_topo;
  PaCO::PacoTopology_t* _server_topo;

  // Save scheduling 
  typedef map<long, paco_comSchedule*> _schedule_info_map_t;
  _schedule_info_map_t _schedule_info_map;    // the map of predefined schedules

};

// Abstract fabrique class for communication scheduling
class paco_fabrique_comScheduling
{
public:
  virtual ~paco_fabrique_comScheduling() {}
  virtual paco_comScheduling* create_comScheduling()=0;
};

// The scheduling fabrique
typedef paco_fabrique_manager_tpl<paco_fabrique_comScheduling> paco_fabrique_manager_comScheduling;

#endif
