#ifndef DISTRIBUTION_LIBRARY_IS_DEFINED
#define DISTRIBUTION_LIBRARY_IS_DEFINED

#include "FabriqueManagerTpl.h"
#include <PaCO++.h>
#include <string>
#include <map>
#include <vector>

using namespace std;

// To delete ?
struct PieceToSend
{
  unsigned sourceNode; // Quel noeud envoi
  unsigned destNode; // Quel est le noeud destinataire
  unsigned size; // nb octet ! la taille que l'on envoi
  void* id; // Cette identifiant est ensuite redonne a la bibliothque lorsque l'on demande 
            // rellement les donnes
};

typedef vector<PieceToSend> vPieceToSend_t;

class DistributionLibrary {

public:

  DistributionLibrary() {}
  virtual ~DistributionLibrary() {}

  // Topology
  virtual void setSourceTopology(PaCO::PacoTopology_t topo) = 0; // paco_operation
  virtual PaCO::PacoTopology_t getSourceTopology() = 0;          // ??
  virtual void setDestTopology(PaCO::PacoTopology_t topo) = 0;   // paco_operation
  virtual PaCO::PacoTopology_t getDestTopology() = 0;            // ??
  virtual void setNodeRank(long Rank) = 0;                       // paco_operation
  virtual long getNodeRank() = 0;                                // by user ?

  // Com Id management
  virtual bool setComId(long id) { return false;}
  virtual bool freeComId(long id) { return false;}

  // If the library is parallel
  virtual void  setCommunicator(void * group) = 0;
  //  virtual void* getCommunicator() = 0; PROPOSED EXTENSION

  // Client Side
  // Computes communications vector (or eventually matrix)
  // Cette mthode est appelle par PaCO++ pour demander la matrice de communication
  virtual PieceToSend* computePiecesToSend(unsigned& size) = 0;
  // virtual PieceToSend* computePiecesToSend2(unsigned& size) = 0; SHOULD COMPUTE DIRECTLY THE FULL MATRIX
  // Sets client-side data 
  // Pointeur vers les donnes client
  virtual void setDataPtr(void* dataPtr) = 0;

  // Get data
  // PaCO++ demande une quantit size pour le server_node
  // length est la taille relle que l'on lui donne
  // end permet de fragmenter le message en plusieurs buffer : si on veut envoyer 
  // qu'un seul buffer on repond true sinon on repond false
  // pid est l'id contenu dans PieceToSend
  // size est la taille que PaCO++ veut et on soustrait ce qu'on donne rellement
  virtual void* getClientData(void *pid, int server_node, long & size, long & length, bool & end) = 0;
  // PaCO++ demande le descripteur qui sera attach  l'envoi des donnes.
  // PaCO++ caste le void * dans le type dcrit dans le fichier de configuration de la lib 
  // situ dans le rpertoires Types.
  virtual void* getDescr() = 0;

  // Free
  virtual void clientFree() = 0;

  // Server Side
  // Sets server-side data return true if data complete
  // element_nb : nombre d'etls de la sequence != 1 si
  // si le message est fragment
  virtual bool insertData(void* rcptBuf, unsigned long element_nb) = 0; 
  virtual void setDescr(void * descr) = 0;

  // getData for the servant
  virtual void* getServerData(long & length) = 0;

  // Free
  virtual void serverFree() = 0;
};

// Abstract class to be implemeted by each fabrique
class paco_fabrique_distribution {
public:
  virtual ~paco_fabrique_distribution() {}
  virtual DistributionLibrary* create()=0;
};

// The distribution fabrique
typedef paco_fabrique_manager_tpl<paco_fabrique_distribution> paco_fabrique_manager_distribution;

#endif
