#ifndef PACO_PLUSPLUS_OPERATION_H_IS_DEFINED
#define PACO_PLUSPLUS_OPERATION_H_IS_DEFINED

#ifdef PADICO
#include <Padico/Timing.h>
#endif

#include "PaCO++_types.h"

#include <string>
#include <map>
#include <iostream>
#include <vector>

class paco_mutex;
class paco_condition;
class paco_com;
class DistributionLibrary;
class paco_comScheduling;

class paco_fabrique_manager;
class paco_fabrique_thread;
class paco_fabrique_com;
class paco_fabrique_comScheduling;

#define PACO_WARN_DISTLIB

// This class represent a context for a parallel operation
class PaCO_operation
{
 public:
  PaCO_operation(int _number_arguments);
  virtual ~PaCO_operation();
  
  /* External API */

  /** Associate a pfm to this object. By default, paco_operation is associated to a FabriqueManager (cf paco_getFabriqueManager()) */
  virtual void   setFabManager(paco_fabrique_manager* fm);

  /* Advanced libs part **/
  /** If not set, use the fabrique manager */ 
  virtual void setComFab(paco_fabrique_com* fc);
  /** If not set, use the fabrique manager */ 
  virtual void setLibCom(const string& LibCom, void * _group);

  /** If not set, use the fabrique manager */ 
  virtual void setThreadFab(paco_fabrique_thread* ft);
  /** If not set, use the fabrique manager */ 
  virtual void setLibThread(const string& LibThread);
  /** If not set, use the fabrique manager */ 

  /** If not set, use the fabrique manager */ 
  virtual void setComSchedulingFab(paco_fabrique_comScheduling* ft);
  /** If not set, use the fabrique manager */ 
  virtual void setLibComScheduling(const string& LibComScheduling);
  /** If not set, use the fabrique manager */ 

  /** Arg part **/
  virtual void   setDisLibArg(int arg_number, const string& lib_name, const string& way) throw(BadWayString);
  // Deprecated
  // virtual void setTypeArg(int arg_number, Pattron * temp, const string& lib_name, const string& way) throw(BadWayString);
  // virtual void   setMode(PaCO::distLoc_t mode, int param, const string& way) throw(BadWayString);
  // virtual void   initArg(PaCO::PacoInitData_t data, int nb_param, const string& way) throw(BadWayString);
  // virtual int    setDistribConfig(void * config, int number, const string& way) throw(BadWayString, BadWaySetStringDistribConfig);
  // virtual void * getDistribConfig(int number, const string& way) throw(BadWayString);

  /** Object part **/
  virtual void setTypeClient(bool type);
  virtual void setClientTopo(PaCO::PacoTopology_t _client_topo, const string& way) throw(BadWayString);
  virtual void setClientTopo(PaCO::PacoTopology_t _client_topo);
  virtual void setServerTopo(PaCO::PacoTopology_t _serveur_topo, const string& way) throw(BadWayString);
  virtual void init(int rank, int total);  

  // Model -> communication matrix
  // Effective Schedule is hidden (reset by clearSchedule & set_beta/k)
  virtual void noCommunicationMemory(); // by default: compute communication & schedule info
  virtual void useCommunicationMemoryId(long id) throw (InvalidArgument); // save communication & scheduling info if it does not exist, use it otherwise -- SHOULD be >= 0
  virtual void clearCommunicationMemoryId(long id) throw (InvalidArgument); // clear this entry, nothing if unknown, exception if invalid (<0)
  virtual void clearScheduleMemoryId(long id) throw (InvalidArgument); // clear this entry, nothing if unknown, exception if invalid (<0)

  // Internal API
  virtual void setClientId(CORBA::Short request_uid);
  virtual void init_return_context(PaCO_operation * op);
  virtual void init_context(PaCO_operation * op);
  virtual void init_context_args(PaCO_operation * op);
  virtual void init_context_proxy(PaCO_operation * op);
  virtual void init_complete_context(PaCO_operation * op) { cerr << "WARNING : init_complete_context of PaCO_Operation" << endl;}
  virtual void configureTopo();
  
 public:
  int return_op;
  int number_arguments;
  int myRank;
  PaCO::PacoTopology_t mytopo;
  int number_stubs;
  PaCO::distLoc_t _mode;
  // Fabrique Manager
  paco_fabrique_manager * FabManager;
  // Thread
  paco_fabrique_thread * FabThread;
  paco_mutex * my_mutex;
  paco_condition * my_condition;
  paco_mutex * my_mutex2;
  paco_condition * my_condition2;
  paco_mutex * my_mutex3;
  // Communication
  void * group;
  paco_com * my_com;
  paco_fabrique_com * FabCom;
  // Distribution
  DistributionLibrary ** libraries_in;
  DistributionLibrary ** libraries_out;

#ifdef PACO_WARN_DISTLIB
  std::vector<char*> libraries_in_name;
  std::vector<char*> libraries_out_name;
#endif

  //ComSchedulr
  paco_comScheduling* my_comScheduling; // only for the the whole interface !!!
  paco_fabrique_comScheduling* FabComScheduling; 
  // Misc
  bool typeClient; // false : sequentiel, true : parallel defaut : false
  PaCO::PacoTopology_t serveur_topo_aller;
  PaCO::PacoTopology_t serveur_topo_retour;
  bool termine;
  bool termine2;
  CORBA::Short client_id;

  // Exception handling
  bool _exception;
  int _level;

  // SHOULD BE PRIVATE !!!

  long _current_id; // -1 means no not use schedule nor scheduling memory

  // Save communication schedule
  typedef struct {
    unsigned** com_matrix; // matrix of unsigned
    void***    info_vector; // vector of void**
  } _com_info_t;
  typedef map<long, _com_info_t*> _com_info_map_t;
  _com_info_map_t _com_info_map;     // The map of predefined communication matrices
  
  void          clearAllScheduleMemory();
  _com_info_t*  getComMemoryId(long id);


#ifdef PADICO
  // For internal Timing: Padico_Timing
  padico_timing_t *_ticks;
#endif
};

#endif
