#ifndef PACO_PLUSPLUS_SRC_H_IS_DEFINED
#define PACO_PLUSPLUS_SRC_H_IS_DEFINED

#include <PaCO++.h>
#include <vector>
#include <string>
#include <iostream>

#include <FabriqueManager.h>

using namespace std;

void* paco_orb_run(void* arg);

paco_fabrique_manager* paco_getFabriqueManager();

class InterfaceManager_impl: virtual public POA_PaCO::InterfaceManager
{
public:
    // Constructors
    InterfaceManager_impl(CORBA::ORB_ptr current_orb);
    virtual ~InterfaceManager_impl();

    // Methods of InterfaceManager
    virtual void setConnectionInfos();
    virtual void setTotalNode(CORBA::Short totalNode); /** obsolete **/
    virtual void setTopo(const PaCO::PacoTopology_t& topo); 
    virtual void setNewNode(const char * node, const CORBA::Short rank);
    virtual void setWorkNode(CORBA::Short totalNode);
    virtual void setReturn(const char * ref);
    virtual char * getReturn();
    virtual void return_ok(const CORBA::Short client_id);

    // Permet aux noeuds d'un client d'avoir le meme client_id !
    virtual CORBA::Short getClientId() = 0;

    // Methods of ParallelKernel
    virtual PaCO::ConnectionInfos* getConnectionInfos();
    // Method to generate to update context when the manager is configured
    virtual void updateContexts() = 0;

    void setReturn(int rtn);
    virtual void finish() = 0;

protected:
    CORBA::ORB_ptr _orb;
    PaCO::ConnectionInfos::_nodes_seq _nodes;
    CORBA::Short _current_uid;
    CORBA::Short _totalNode;
    CORBA::Short _totalNodeReceived;
    CORBA::Short _workNode;
    PaCO::ConnectionInfos _informations;
    PaCO::ConnectionInfos _informations_temp;
    PaCO::PacoTopology_t serveur_topo;
    string return_str;
    int _rtn; // To known if the interface manager is a 
                // return interface manager
};

class InterfaceParallel_impl: virtual public POA_PaCO::InterfaceParallel
{
public:
    // Constructors
    InterfaceParallel_impl(CORBA::ORB_ptr current_orb, char * ior);
    virtual ~InterfaceParallel_impl();

    // Methods of InterfaceParallel
    CORBA::Short getMyRank();
    CORBA::Short getTotalNode();
    CORBA::Short getDeployRank();
    void init_InterfaceParallel(CORBA::Short myRank, CORBA::Short totalNode);
    void deploy(CORBA::Short rank);
    // Methods of ParallelKernel
    virtual PaCO::ConnectionInfos* getConnectionInfos();
    // Return object
    void refReturnObject(const char * ref);

    // EXCEPTION !
    virtual void setRef(const PaCO::InterfaceParallel::ref & ref_nodes);
    // Actually implemented by the PaCO++ generated layer
    virtual void addException(const CORBA::Short except) {}

    PaCO::InterfaceParallel::ref _ref_nodes;

protected:
    CORBA::ORB_ptr _orb;
    string return_object;

private:
    CORBA::Short _deployRank;
    CORBA::Short _myRank;
    CORBA::Short _totalNode;
    string _ior;
    

};

#endif
